@@ -2,7 +2,7 @@
  * Unsquash a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2009, 2010, 2013, 2019, 2021, 2022
+ * Copyright (c) 2009, 2010, 2013, 2019, 2021, 2022, 2023
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -209,7 +209,6 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 	i.mode = lookup_type[header.base.inode_type] | header.base.mode;
 	i.type = header.base.inode_type;
-	i.time = sBlk.s.mkfs_time;
 	i.inode_number = inode_number++;
 
 	switch(header.base.inode_type) {
@@ -232,7 +231,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 			i.data = inode->file_size;
 			i.offset = inode->offset;
 			i.start = inode->start_block;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			break;
 		}
 		case SQUASHFS_LDIR_TYPE: {
@@ -254,7 +256,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 			i.data = inode->file_size;
 			i.offset = inode->offset;
 			i.start = inode->start_block;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			break;
 		}
 		case SQUASHFS_FILE_TYPE: {
@@ -274,7 +279,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 					"inode %lld:%d\n", start, offset);
 
 			i.data = inode->file_size;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			i.frag_bytes = inode->fragment == SQUASHFS_INVALID_FRAG
 				?  0 : inode->file_size % sBlk.s.block_size;
 			i.fragment = inode->fragment;
@@ -316,6 +324,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 					"inode symbolic link %lld:%d\n", start, offset);
 			i.symlink[inodep->symlink_size] = '\0';
 			i.data = inodep->symlink_size;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 		}
  		case SQUASHFS_BLKDEV_TYPE:
@@ -336,11 +348,19 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 					"inode %lld:%d\n", start, offset);
 
 			i.data = inodep->rdev;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 			}
 		case SQUASHFS_FIFO_TYPE:
 		case SQUASHFS_SOCKET_TYPE:
 			i.data = 0;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 		default:
 			EXIT_UNSQUASH("Unknown inode type %d in "
@@ -478,7 +498,7 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 	}
 
 	if(needs_sorting)
-		sort_directory(dir);
+		sort_directory(&(dir->dirs), dir->dir_count);
 
 	/* check directory for duplicate names and sorting */
 	if(check_directory(dir) == FALSE) {
